------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------



with Config;
use Config;
with Project;
use Project;
with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;
with Double_Util;
use Double_Util;
with Translate;
use Translate;
with Tasks;
use Tasks;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;
with Unbounded_Strings;
use Unbounded_Strings;


package body Call_Scheduling_Stuff is


   procedure Build_And_Check_Scheduling (
         A_Processor :        Processor_Ptr;      
         Message     : in out Unbounded_String;   
         Period      : in     Natural;            
         Sched       :    out Scheduling_Sequence ) is 

      Check : Unbounded_String := Empty_String;  
   begin

      Sort(Sys.Tasks, Increasing_Name'access);

      Reset(Sched);
      Build_Scheduling_Sequence(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name, Sched, Period);
      Message:=Message & Lb_Number_Of_Preemption(Current_Language) & Sched.
         Number_Of_Preemption'Img & Unbounded_Lf;

      Check_Temporal_Constraints (Sys.Tasks, Sched, A_Processor.Name,
         Check);
      if Check = Empty_String
            then
         Message:=Message & Lb_No_Missed_Deadline(Current_Language) &
            Unbounded_Lf;
      else
         Message:=Message & Check & Unbounded_Lf;
      end if;

   exception
      when Scheduler.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_15(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_15(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Task_Must_Have_Period_Equal_To_Deadline=>
         Message:=Message & Lb_Compute_Error_14(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Task_Model_Error =>
         Message:=Message & Lb_Compute_Error_1(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Priority_Error =>
         Message:=Message & Lb_Compute_Error_11(Current_Language) &
            Unbounded_Lf;


   end Build_And_Check_Scheduling;






   procedure Compute_Scheduling_Period (
         A_Processor :        Processor_Ptr;   
         Message     : in out Unbounded_String ) is 
      Period  : Natural;  
      Result  : Double;  
      Free_Ut : Double;  
   begin
      Period:=Scheduling_Period(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name);
      Message:=Message & Lb_Scheduling_Period(Current_Language) & Period'
         Img  &
         To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 6). ") & Unbounded_Lf;

      Result:=Processor_Utilization_Over_Period(A_Processor.Scheduler.all,
         Sys.Tasks,
         A_Processor.Name);

      Free_Ut:=Double(Period)*(1.0-Result);
      Free_Ut:=Max(0.0,Free_Ut);
      Message:=Message & Lb_Minus
         & Format(Free_Ut) &
         Lb_Free_Unit_Time(Current_Language) & Unbounded_Lf;


   exception
      when Scheduler.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_2(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Task_Model_Error =>
         Message:=Message & Lb_Compute_Error_3(Current_Language) &
            Unbounded_Lf;


   end Compute_Scheduling_Period;





   procedure Compute_Response_Time (
         A_Processor :        Processor_Ptr;   
         Message     : in out Unbounded_String ) is 
      Rt : Response_Table;  

      Msg : Unbounded_String := Empty_String;  

   begin

      Initialize(Rt);

      Compute_Response_Time(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name, Msg, Rt);
      Message:=Message & Lb_Task_Response_Time(Current_Language) & Msg &
         Unbounded_Lf;



      for I in 0..Response_Range(Get_Number_Of_Task_From_Processor(Sys.
               Tasks,A_Processor.Name)-1) loop
         if (Rt(I).Task_Item /= null)
               then
            Message:=Message &  To_Unbounded_String("    ") & Rt(I).
               Task_Item.Name & To_Unbounded_String(" => ") & Format(Rt(I).
               Data) &
               Unbounded_Lf;
         end if;
      end loop;


   exception
      when Scheduler.Task_Must_Have_Period_Equal_To_Deadline=>
         Message:=Message & Lb_Compute_Error_4(Current_Language) &
            Unbounded_Lf;
      when Task_Set.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_5(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_5(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Task_Model_Error =>
         Message:=Message & Lb_Compute_Error_6(Current_Language) &
            Unbounded_Lf;
      when Processor_Utilization_Exceeded =>
         Message:=Message & Lb_Compute_Error_7(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Start_Time_Error =>
         Message:=Message & Lb_Compute_Error_12(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Offset_Error =>
         Message:=Message & Lb_Compute_Error_13(Current_Language) &
            Unbounded_Lf;

   end Compute_Response_Time;



   procedure Processor_Utilization (
         A_Processor :        Processor_Ptr;   
         Message     : in out Unbounded_String ) is 

      Result : Double;  
      Msg    : Unbounded_String;  
   begin

      Result:=Processor_Utilization_Over_Deadline(A_Processor.Scheduler.all,
         Sys.Tasks,
         A_Processor.Name);
      Message:= Message & Lb_Utilization_With_Deadline(Current_Language) &
         Format(Result) & Lb_Point & Unbounded_Lf;

      Result:=Processor_Utilization_Over_Period(A_Processor.Scheduler.all,
         Sys.Tasks,
         A_Processor.Name);
      Message:= Message & Lb_Utilization_With_Period(Current_Language) &
         Format(Result) &
         To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 6). ") & Unbounded_Lf;

      Bound_On_Processor_Utilization(A_Processor.Scheduler.all, Sys.Tasks,
         A_Processor.Name, Result, Msg);
      Message:= Message & Lb_Utilization_Bound(Current_Language) & Format(
         Result) &
         Msg;


   exception
      when Scheduler.Task_Must_Be_Periodic=>
         Message:=Message & Lb_Compute_Error_9(Current_Language) &
            Unbounded_Lf;
      when Scheduler.Task_Model_Error =>
         Message:=Message & Lb_Compute_Error_10(Current_Language) &
            Unbounded_Lf;


   end Processor_Utilization;






end Call_Scheduling_Stuff;




