------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Text_Io;
use Text_Io;


with Fixed_Strings;
use Fixed_Strings;

with Dependencies;
use Dependencies;


use Buffer_Set.Generic_Buffer_Set;


package body Buffer_Set is



   procedure Fill_Descriptor (
         My_Buffers : in     Buffers_Set;       
         Desc       :    out Buffer_Descriptor; 
         A_Buffer   : in     Buffer_Ptr         ) is 

   begin
      Desc.Name := To_Bounded_String(To_String(A_Buffer.Name));
      Desc.Location:= To_Bounded_String(To_String(A_Buffer.Location));
      Desc.Size:= A_Buffer.Size;
      Desc.Rules:=A_Buffer.Rules;

   end Fill_Descriptor;


   procedure Extract_Descriptor (
         My_Buffers : in     Buffers_Set;       
         Desc       : in     Buffer_Descriptor; 
         A_Buffer   :    out Buffer_Ptr         ) is 

      New_Buffer : Buffer_Ptr;  

   begin
      New_Buffer := new Buffer;
      A_Buffer.Name:=To_Unbounded_String(To_String(Desc.Name));
      A_Buffer.Location:=To_Unbounded_String(To_String(Desc.Location));
      A_Buffer.Size:=Desc.Size;
      A_Buffer.Rules:=Desc.Rules;

   end Extract_Descriptor;



   procedure Add_Buffer (
         My_Buffers : in out Buffers_Set;       
         Name       : in     Unbounded_String;  
         Size       : in     Natural;           
         Location   : in     Unbounded_String;  
         Rules      : in     Buffer_Rules_Table ) is 


      Desc        : Buffer_Descriptor;  
      A_Buffer    : Buffer_Ptr;  
      My_Iterator : Iterator;  

   begin


      if (Name = "")
            then
         raise Invalid_Parameter;
      end if;

      if (Location = "")
            then
         raise Invalid_Parameter;
      end if;

      if (Size = 0)
            then
         raise Invalid_Parameter;
      end if;

      if(Get_Number_Of_Elements(My_Buffers) > 0)
            then

         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Buffers, A_Buffer, My_Iterator);
            if (Name = A_Buffer.Name)
                  then
               raise Buffer_Already_Defined;
            end if;

            exit when Is_Last_Element(My_Buffers, My_Iterator) = True;

            Next_Element(My_Buffers, My_Iterator);
         end loop;
      end if;


      Desc.Name:=To_Bounded_String(To_String(Name));
      Desc.Location:=To_Bounded_String(To_String(Location));
      Desc.Size:=Size;
      Desc.Rules:=Rules;

      Extract_Descriptor(My_Buffers, Desc, A_Buffer);


      Add(My_Buffers,A_Buffer);


   end Add_Buffer;



   procedure Dump (
         My_Buffers : in     Buffers_Set ) is 

      A_Buffer    : Buffer_Ptr;  
      My_Iterator : Iterator;  

   begin
      Put_Line ("Number of buffers : " & Get_Number_Of_Elements(
            My_Buffers)'
         Img);

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Buffers,A_Buffer,My_Iterator);
         Dump(A_Buffer.all);
         New_Line;

         exit when Is_Last_Element(My_Buffers, My_Iterator) = True;

         Next_Element(My_Buffers, My_Iterator);

      end loop;
   end Dump;


   function Get (
         My_Buffers  : in     Buffers_Set;      
         Buffer_Name : in     Unbounded_String; 
         Param_Name  : in     Buffer_Parameters ) 
     return Natural is 

      A_Buffer    : Buffer_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( Param_Name /= Size)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         if (A_Buffer.Name = Buffer_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Buffers, My_Iterator) = True;

         Next_Element(My_Buffers, My_Iterator);

      end loop;

      return A_Buffer.Size;

   end Get;



   function Get (
         My_Buffers  : in     Buffers_Set;      
         Buffer_Name : in     Unbounded_String; 
         Param_Name  : in     Buffer_Parameters ) 
     return Unbounded_String is 

      A_Buffer    : Buffer_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( Param_Name /= Location)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         if (A_Buffer.Name = Buffer_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Buffers, My_Iterator) = True;

         Next_Element(My_Buffers, My_Iterator);

      end loop;

      return A_Buffer.Location;

   end Get;


   procedure Set (
         My_Buffers  : in out Buffers_Set;       
         Buffer_Name : in     Unbounded_String;  
         Param_Name  : in     Buffer_Parameters; 
         Param_Value : in     Natural            ) is 
      A_Buffer    : Buffer_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Size)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         if (A_Buffer.Name = Buffer_Name)
               then
            A_Buffer.Size:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Buffers, My_Iterator) = True;

         Next_Element(My_Buffers, My_Iterator);

      end loop;
   end Set;


   procedure Set (
         My_Buffers  : in out Buffers_Set;       
         Buffer_Name : in     Unbounded_String;  
         Param_Name  : in     Buffer_Parameters; 
         Param_Value : in     Unbounded_String   ) is 
      A_Buffer    : Buffer_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Location)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Buffers,A_Buffer,My_Iterator);

         if (A_Buffer.Name = Buffer_Name)
               then
            A_Buffer.Location:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Buffers, My_Iterator) = True;

         Next_Element(My_Buffers, My_Iterator);

      end loop;
   end Set;



end Buffer_Set;
