------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Config;
use Config;

with Translate;
use Translate;

with Tasks;
use Tasks;

with Scheduler;
use Scheduler;
use Scheduler.Double_Tasks_Parameters;

with Double_Util;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;


with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Unbounded_Strings;
use  Unbounded_Strings;

package body Scheduler.Dm is


   procedure Initialize (
         A_Scheduler : in out Dm_Scheduler ) is 
   begin
      Reset(A_Scheduler);
   end Initialize;

   procedure Finalize (
         A_Scheduler : in out Dm_Scheduler ) is 
   begin
      null;
   end Finalize;

   procedure Adjust (
         A_Scheduler : in out Dm_Scheduler ) is 
   begin
      null;
   end Adjust;

   procedure Dispatched_Reset (
         A_Scheduler : in out Dm_Scheduler ) is 
   begin
      A_Scheduler.Name:= Deadline_Monotonic;
   end Dispatched_Reset;


   function Dispatched_Build_Tcb (
         My_Scheduler : in     Dm_Scheduler; 
         A_Task       :        Task_Ptr      ) 
     return Tcb_Ptr is 

      A_Tcb : Tcb_Ptr;  
   begin
      A_Tcb:=new Dm_Tcb;

      A_Tcb.Tsk:=A_Task;
      A_Tcb.Activation:=1;
      A_Tcb.Wake_Up_Time:=A_Task.Start_Time;
      A_Tcb.Rest_Of_Capacity:=A_Task.Capacity;

      if(A_Task.Deadline /= 0)
            then
         Dm_Tcb_Ptr(A_Tcb).Static_Deadline:= A_Task.Deadline;
      else
         Dm_Tcb_Ptr(A_Tcb).Static_Deadline:=Periodic_Task_Ptr(
            A_Task).Period;
      end if;

      return A_Tcb;
   end Dispatched_Build_Tcb;


   procedure Dispatched_Check_Before_Scheduling (
         My_Scheduler   : in     Dm_Scheduler;    
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

   begin

      Periodic_Control(My_Scheduler, My_Tasks, Processor_Name);

   end Dispatched_Check_Before_Scheduling;


   procedure Dispatched_Do_Election (
         My_Scheduler : in     Dm_Scheduler; 
         A_Tcb        : in     Tcb_Table;    
         At_Time      :        Natural;      
         Elected      : in out Tasks_Range;  
         No_Task      : in out Boolean       ) is 

      Smallest_Static_Deadline : Natural     := Natural'Last;  
      I                        : Tasks_Range := 0;  
   begin

      loop
         if (A_Tcb(I).Wake_Up_Time <= At_Time) and
               (Dm_Tcb_Ptr(A_Tcb(I)).Static_Deadline <
               Smallest_Static_Deadline) and
               Check_Offset(A_Tcb(I), At_Time) and
               (A_Tcb(I).Rest_Of_Capacity /= 0)
               then
            Smallest_Static_Deadline:=Dm_Tcb_Ptr(A_Tcb(I)).
               Static_Deadline;
            Elected:=I;
         end if;

         I:=I+1;
         exit when A_Tcb(I)=null;
      end loop;

      if Smallest_Static_Deadline= Natural'Last
            then
         No_Task:=True;
      else
         No_Task:=False;
      end if;

   end Dispatched_Do_Election;



   procedure Set_Priority_According_To_Dm (
         My_Scheduler   : in     Dm_Scheduler;    
         My_Tasks       : in out Tasks_Set;       
         Processor_Name : in     Unbounded_String ) is 

      Iterator1    : Iterator;  
      Task1        : Task_Ptr;  
      Current_Prio : Priority_Range := High_Priority;  

   begin

      Sort(My_Tasks, Increasing_Deadline'access);

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);

         if (Task1.Location = Processor_Name)
               then
            Task1.Priority:=Current_Prio;
            Current_Prio:=Current_Prio-1;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Set_Priority_According_To_Dm;


   function Compute_Workload (
         My_Tasks     : in     Tasks_Set; 
         Current_Task : in     Task_Ptr;  
         Q            : in     Integer    ) 
     return Double is 

      Iterator2 : Iterator;  
      Task2     : Task_Ptr;  
      Rn,  
      Rn1       : Double;  

   begin

      Rn:=-0.1;
      Rn1:=0.0;

      while Rn1 /= Rn  loop

         Reset_Iterator(Iterator2);

         Rn:=Rn1;
         Rn1:=Double(Q*Current_Task.Capacity+Current_Task.Blocking_Time);

         loop
            Current_Element(My_Tasks, Task2, Iterator2);

            if (Task2.Priority> Current_Task.Priority)
                  then
               Rn1:=Rn1+
                  Double_Util.Ceiling((Rn+Double(Periodic_Task_Ptr(Task2).
                        Jitter))/Double(Periodic_Task_Ptr(Task2).Period))*
                  Double(Task2.Capacity) ;
            end if;

            exit when Is_Last_Element(My_Tasks, Iterator2) = True;
            Next_Element(My_Tasks, Iterator2);

         end loop;

      end loop;

      return Rn1;

   end Compute_Workload;


   procedure Dispatched_Response_Time (
         My_Scheduler   : in     Dm_Scheduler;     
         My_Tasks       : in out Tasks_Set;        
         Processor_Name : in     Unbounded_String; 
         Msg            : in out Unbounded_String; 
         Response_Time  :    out Response_Table    ) is 


      Tmp       : Tasks_Set;  
      Iterator1 : Iterator;  
      Task1,  
      Task_Tmp  : Task_Ptr;  
      Number    : Response_Range;  
      Wi,  
      Ri        : Double         := 0.0;  
      K         : Integer;  

   begin



      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[2], page 3, ") & Lb_Equation(
         Current_Language)
         & To_Unbounded_String("4). ");




      -- check if tasks are periodics 
      Periodic_Control(My_Scheduler, My_Tasks, Processor_Name);


      -- Check processor utilization first
      if (Processor_Utilization_Over_Period(My_Scheduler, My_Tasks,
               Processor_Name) > 1.0)
            then
         raise Processor_Utilization_Exceeded;
      end if;

      -- check start time
      Start_Time_Control(My_Scheduler, My_Tasks, Processor_Name);

      -- check offset
      Offset_Control(My_Scheduler, My_Tasks, Processor_Name);


      -- We will change priority task to compute response time
      -- so, we will use a temporary task set

      Reset_Iterator(Iterator1);
      loop


         Current_Element(My_Tasks, Task1, Iterator1);
         if (Task1.Location = Processor_Name)
               then
            Task_Tmp:= new Periodic_Task;
            Task_Tmp.all:=Task1.all;
            Add(Tmp,Task_Tmp);
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;


      -- Now, compute response time

      Set_Priority_According_To_Dm(My_Scheduler, Tmp, Processor_Name);
      Reset_Iterator(Iterator1);
      Number:=0;
      K:=1;

      loop

         Current_Element(Tmp, Task1, Iterator1);


         Response_Time(Number).Data:=0.0;
         Response_Time(Number).Task_Item:=Task1;


         -- compute response time for a given task
         --
         loop
            Wi:=Compute_Workload(Tmp, Task1, K);

            Ri:= Wi + Double(Periodic_Task_Ptr(Task1).Jitter) - Double(
               Periodic_Task_Ptr(Task1).Period) *Double(K-1);
            if(Ri > Response_Time(Number).Data)
                  then
               Response_Time(Number).Data:=Ri;
            end if;

            exit when Ri <= Double(Periodic_Task_Ptr(Task1).Period);

            K:=K+1;

         end loop;

         Number:=Number+1;

         exit when Is_Last_Element(Tmp, Iterator1) = True;
         Next_Element(Tmp, Iterator1);


      end loop;


   end Dispatched_Response_Time;


   procedure Dispatched_Bound_On_Processor_Utilization (
         My_Scheduler   : in     Dm_Scheduler;     
         My_Tasks       : in     Tasks_Set;        
         Processor_Name : in     Unbounded_String; 
         Result         : in out Double;           
         Msg            : in out Unbounded_String  ) is 

      Nb : Double := Double (Get_Number_Of_Task_From_Processor (My_Tasks,
        Processor_Name));  
   begin
      Msg:=To_Unbounded_String(" (") & Lb_See(Current_Language) &
         To_Unbounded_String("[1], page 16, ") & Lb_Theorem(
         Current_Language) &
         To_Unbounded_String("8). ") & Unbounded_Lf;

      Result:=Nb*( Pow(2.0,(1.0/Nb)) - 1.0);
   end Dispatched_Bound_On_Processor_Utilization;





end Scheduler.Dm ;


