------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------




with Text_Io;
use Text_Io;

with Tasks;
use Tasks;

with Natural_Util;
use Natural_Util;

with Config;
use Config;


package body Scheduler is




   procedure Set_Preemptive (
         My_Scheduler : in out Generic_Scheduler'Class; 
         Preempt      : in     Preemptive_Type          ) is 
   begin
      My_Scheduler.Is_Preemptive:=Preempt;
   end Set_Preemptive;


   function Get_Preemptive (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Preemptive_Type is 
   begin
      return My_Scheduler.Is_Preemptive;
   end Get_Preemptive;


   procedure Dump (
         My_Scheduler : in     Generic_Scheduler'Class ) is 
   begin
      Put_Line("Dump scheduler :");
      Put_Line("scheduler name : " & My_Scheduler.Name'Img);
      Put_Line("is preemptive : " & My_Scheduler.Is_Preemptive'Img);
   end Dump;


   procedure Initialize (
         A_Scheduler : in out Generic_Scheduler ) is 
   begin
      Reset(A_Scheduler);
   end Initialize;


   procedure Finalize (
         A_Scheduler : in out Generic_Scheduler ) is 
   begin
      null;
   end Finalize;


   procedure Adjust (
         A_Scheduler : in out Generic_Scheduler ) is 
   begin
      null;
   end Adjust;

   procedure Initialize (
         A_Tcb_Table : in out Tcb_Table ) is 
   begin
      for I in Tasks_Range loop
         A_Tcb_Table(I):=null;
      end loop;
   end Initialize;



   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Unbounded_String is 
   begin
      return To_Unbounded_String(My_Scheduler.Name'Img);
   end Get_Name;

   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Scheduler_Type is 
   begin
      return My_Scheduler.Name;
   end Get_Name;



   function Scheduling_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Natural is 

      Period : Natural := 1;  

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin


      -- check if tasks are periodics
      Periodic_Control(My_Scheduler, My_Tasks, Processor_Name);


      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks, A_Task, My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            Period:= Natural_Util.Lcm(Period,
               Periodic_Task_Ptr(A_Task).Period);
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);
      end loop;


      return Period;

   end Scheduling_Period;


   function Processor_Utilization_Over_Deadline (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double is 

      Utilization : Double := 0.0;  

      My_Iterator : Iterator;  
      A_Task      : Task_Ptr;  

   begin

      -- check if tasks are periodics
      Periodic_Control(My_Scheduler, My_Tasks, Processor_Name);


      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            Utilization:=Utilization+(Double(A_Task.Capacity) /
               Double(A_Task.Deadline));
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;


      return Utilization;

   end Processor_Utilization_Over_Deadline;


   function Processor_Utilization_Over_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double is 

      Utilization : Double := 0.0;  

      My_Iterator : Iterator;  
      A_Task      : Task_Ptr;  

   begin

      -- check if tasks are periodics
      Periodic_Control(My_Scheduler, My_Tasks, Processor_Name);



      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location = Processor_Name)
               then
            Utilization:=Utilization+(Double(A_Task.Capacity) /
               Double(Periodic_Task_Ptr(A_Task).Period));
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return Utilization;

   end Processor_Utilization_Over_Period;



   procedure Reset (
         A_Scheduler : in out Generic_Scheduler'Class ) is 
   begin
      A_Scheduler.Is_Preemptive:=Preemptive;
      Dispatched_Reset(A_Scheduler);
   end Reset;


   procedure Build_Scheduling_Sequence (
         My_Scheduler    : in     Generic_Scheduler'Class;     
         My_Tasks        : in     Tasks_Set;                   
         Processor_Name  : in     Unbounded_String;            
         Result          :    out Scheduling_Sequence;         
         Given_Last_Time :        Natural                 := 0 ) is 


      Tcbs            : Tcb_Table;  
      My_Iterator     : Iterator;  
      A_Task          : Task_Ptr;  
      Number_Of_Tasks : Tasks_Range := 0;  

      Last_Time : Natural;  

      Current_Time       : Natural         := 0;  
      Current_Time_Unit  : Time_Unit_Range := 0;  
      Elected            : Tasks_Range     := Tasks_Range'First;  
      Previously_Elected : Tasks_Range     := Tasks_Range'First;  
      No_Task            : Boolean;  


   begin


      Check_Before_Scheduling(My_Scheduler, My_Tasks, Processor_Name);


      -- Compute last scheduled time unit
      --

      if(Given_Last_Time = 0)
            then
         Last_Time:=Natural_Util.Min (Max_Scheduling_Period,
            Scheduling_Period (My_Scheduler, My_Tasks, Processor_Name));
      else
         Last_Time:=Given_Last_Time;
      end if;


      -- First : gather information
      --

      Initialize(Tcbs);
      Reset_Iterator(My_Iterator);

      loop

         Current_Element(My_Tasks, A_Task, My_Iterator);

         if (A_Task.Location = Processor_Name)
               then

            Tcbs(Number_Of_Tasks):=Build_Tcb(My_Scheduler, A_Task);

            Number_Of_Tasks:=Number_Of_Tasks+1;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;
         Next_Element(My_Tasks, My_Iterator);


      end loop;

      -- Second : compute scheduling
      --
      Reset(Result);

      while(Current_Time < Last_Time) loop


         Previously_Elected:=Elected;

         Do_Election(My_Scheduler, Tcbs, Current_Time, Elected, No_Task);

         -- No task to schedule
         --
         if (No_Task = True)
               then
            Current_Time:=Current_Time+1;

         else


            if (My_Scheduler.Is_Preemptive = Preemptive)
                  then
               Result.Scheduled(Current_Time_Unit).Data:=Current_Time;
               Result.Scheduled(Current_Time_Unit).Task_Item:=Tcbs(
                  Elected).Tsk;
               Result.Number_Of_Time_Unit:=Result.Number_Of_Time_Unit + 1;
               Tcbs(Elected).Rest_Of_Capacity:=Tcbs(Elected).
                  Rest_Of_Capacity - 1;


               if (Current_Time_Unit > 0)
                     then

                  if ( (Result.Scheduled(Current_Time_Unit - 1).Task_Item.
                           Name /=
                           Result.Scheduled(Current_Time_Unit).Task_Item.
                           Name)
                        and (Result.Scheduled(Current_Time_Unit - 1).Data+
                           1=
                           Result.Scheduled(Current_Time_Unit).Data)
                        and (Tcbs(Previously_Elected).Rest_Of_Capacity/=0)
                        and (Tcbs(Previously_Elected).Rest_Of_Capacity/=
                           Tcbs(Previously_Elected).Tsk.Capacity)
                        )
                        then
                     Result.Number_Of_Preemption:= Result.
                        Number_Of_Preemption + 1;
                  end if;
               end if;

               Current_Time_Unit:=Current_Time_Unit + 1;
               Current_Time:=Current_Time + 1;

            else
               for I in 0 .. (Tcbs(Elected).Tsk.Capacity - 1)
                     loop
                  Result.Scheduled(Current_Time_Unit).Data:=Current_Time;
                  Result.Scheduled(Current_Time_Unit).Task_Item:=Tcbs(
                     Elected).Tsk;
                  Current_Time_Unit:=Current_Time_Unit + 1;
                  Result.Number_Of_Time_Unit:=Result.Number_Of_Time_Unit+
                     1;
                  Current_Time:=Current_Time + 1;
               end loop;
               Tcbs(Elected).Rest_Of_Capacity:=0;
            end if;

            -- Next activation ?
            --
            if (Tcbs(Elected).Rest_Of_Capacity = 0)
                  then
               if (Tcbs(Elected).Tsk.Task_Type=Periodic_Type)
                     then
                  Tcbs(Elected).Activation:=Tcbs(Elected).Activation+1;
                  Tcbs(Elected).Rest_Of_Capacity:=Tcbs(Elected).Tsk.
                     Capacity;
                  Tcbs(Elected).Wake_Up_Time:=Tcbs(Elected).Wake_Up_Time +
                     Periodic_Task_Ptr(Tcbs(Elected).Tsk).Period;
               end if;
            end if;
         end if;

      end loop;

   end Build_Scheduling_Sequence;



   procedure Compute_Response_Time (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in out Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Msg            : in out Unbounded_String;        
         Response_Time  :    out Response_Table           ) is 
   begin
      Dispatched_Response_Time(My_Scheduler, My_Tasks, Processor_Name, Msg,
         Response_Time);
   end Compute_Response_Time;


   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Result         : in out Double;                  
         Msg            : in out Unbounded_String         ) is 
   begin
      Dispatched_Bound_On_Processor_Utilization(My_Scheduler,My_Tasks,
         Processor_Name, Result, Msg);
   end Bound_On_Processor_Utilization;


   -- Return true is the task "task_id" can be schedule now
   -- according its offsets
   --		
   function Check_Offset (
         A_Tcb   : Tcb_Ptr; 
         At_Time : Natural  ) 
     return Boolean is 
      Target_Time : Integer := 0;  
   begin

      for K in Offset_Range loop
         if A_Tcb.Tsk.Offset(K).Activation=A_Tcb.Activation
               then
            Target_Time:=A_Tcb.Wake_Up_Time+A_Tcb.Tsk.Offset(
               K).Value;
            if At_Time<Target_Time
                  then
               return False;
            end if;
         end if;
      end loop;

      return True;
   end Check_Offset;


   procedure Check_Before_Scheduling (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) is 
   begin
      Dispatched_Check_Before_Scheduling(My_Scheduler, My_Tasks,
         Processor_Name);
   end Check_Before_Scheduling;


   procedure Do_Election (
         My_Scheduler : in     Generic_Scheduler'Class; 
         A_Tcb        : in     Tcb_Table;               
         At_Time      :        Natural;                 
         Elected      : in out Tasks_Range;             
         No_Task      : in out Boolean                  ) is 
   begin
      Dispatched_Do_Election(My_Scheduler, A_Tcb, At_Time, Elected,
         No_Task);
   end Do_Election;


   function Build_Tcb (
         My_Scheduler : in     Generic_Scheduler'Class; 
         A_Task       :        Task_Ptr                 ) 
     return Tcb_Ptr is 
   begin
      return Dispatched_Build_Tcb(My_Scheduler, A_Task);
   end Build_Tcb;


   procedure Periodic_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) is 

      Iterator1 : Iterator;  
      Task1     : Task_Ptr;  

   begin

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if (Task1.Location = Processor_Name)
               then
            if  (Task1.Task_Type /= Periodic_Type)
                  then
               raise Task_Must_Be_Periodic;
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;

   end Periodic_Control;



   procedure Priority_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) is 


      Iterator1 : Iterator;  
      Task1     : Task_Ptr;  
      Iterator2 : Iterator;  
      Task2     : Task_Ptr;  

   begin
      -- No duplicate priority
      --

      Reset_Iterator(Iterator1);


      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if (Task1.Location = Processor_Name)
               then
            Reset_Iterator(Iterator2);

            loop
               Current_Element(My_Tasks, Task2, Iterator2);


               if (Task2.Location = Task1.Location)
                     and (Task1.Name /= Task2.Name)
                     and (Task1.Priority = Task2.Priority)
                     then
                  raise Priority_Error;
               end if;

               exit when Is_Last_Element(My_Tasks, Iterator2) = True;
               Next_Element(My_Tasks, Iterator2);

            end loop;


         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Priority_Control;

   -- check start time
   procedure Start_Time_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) is 

      Iterator1 : Iterator;  
      Task1     : Task_Ptr;  

   begin

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if (Task1.Location = Processor_Name)
               and (Task1.Start_Time /= 0)
               then
            raise Start_Time_Error;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Start_Time_Control;



   -- check offset
   --
   procedure Offset_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) is 

      Iterator1 : Iterator;  
      Task1     : Task_Ptr;  

   begin

      Reset_Iterator(Iterator1);

      loop
         Current_Element(My_Tasks, Task1, Iterator1);


         if (Task1.Location = Processor_Name)
               and (
               (Task1.Offset(1).Activation /= 0) or
               (Task1.Offset(2).Activation /= 0) or
               (Task1.Offset(3).Activation /= 0))
               then
            raise Offset_Error;
         end if;

         exit when Is_Last_Element(My_Tasks, Iterator1) = True;
         Next_Element(My_Tasks, Iterator1);

      end loop;
   end Offset_Control;


end Scheduler ;


