------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Tasks;
use Tasks;

with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;

with Ada.Finalization;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Config;

with Ada.Numerics.Aux;
use Ada.Numerics.Aux;

with Time_Units;
use Time_Units;
use Time_Units.A_Time_Unit;

with Scheduling_Sequences;
use Scheduling_Sequences;

with Tables;

with Double_Util;

package Scheduler is



   type Scheduler_Type is 
         (Earliest_Deadline_First, 
          Least_Laxity_First,      
          Rate_Monotonic,          
          Deadline_Monotonic,      
          Highiest_Priority_First); 

   type Preemptive_Type is 
         (Preemptive,    
          Non_Preemptive); 



   package Double_Tasks_Parameters is new Tables(Double,
      Config.Max_Tasks, Double_Util.Put, Double_Util.Initialize);
   subtype Response_Table is Double_Tasks_Parameters.Table;
   subtype Response_Range is Double_Tasks_Parameters.Table_Range;




   Task_Model_Error                        : exception;  
   Task_Must_Be_Periodic                   : exception;  
   Task_Must_Have_Period_Equal_To_Deadline : exception;  
   Processor_Utilization_Exceeded          : exception;  
   Priority_Error                          : exception;  
   Start_Time_Error                        : exception;  
   Offset_Error                            : exception;  



   type Generic_Scheduler is abstract tagged private; 
   type Scheduler_Ptr is access all Generic_Scheduler'Class; 


   procedure Initialize (
         A_Scheduler : in out Generic_Scheduler ); 


   procedure Finalize (
         A_Scheduler : in out Generic_Scheduler ); 


   procedure Adjust (
         A_Scheduler : in out Generic_Scheduler ); 


   procedure Reset (
         A_Scheduler : in out Generic_Scheduler'Class ); 


   type Tcb is tagged 
      record 
         Tsk              : Task_Ptr;  
         Wake_Up_Time     : Natural;  
         Rest_Of_Capacity : Natural;  
         Activation       : Natural;  
      end record; 

   type Tcb_Ptr is access all Tcb'Class; 
   type Tcb_Table is array (Tasks_Range) of Tcb_Ptr; 

   procedure Initialize (
         A_Tcb_Table : in out Tcb_Table ); 

   procedure Build_Scheduling_Sequence (
         My_Scheduler    : in     Generic_Scheduler'Class;     
         My_Tasks        : in     Tasks_Set;                   
         Processor_Name  : in     Unbounded_String;            
         Result          :    out Scheduling_Sequence;         
         Given_Last_Time :        Natural                 := 0 ); 


   procedure Compute_Response_Time (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in out Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Msg            : in out Unbounded_String;        
         Response_Time  :    out Response_Table           ); 


   procedure Set_Preemptive (
         My_Scheduler : in out Generic_Scheduler'Class; 
         Preempt      : in     Preemptive_Type          ); 


   function Get_Preemptive (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Preemptive_Type; 


   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Scheduler_Type; 

   function Get_Name (
         My_Scheduler : in     Generic_Scheduler'Class ) 
     return Unbounded_String; 

   function Scheduling_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Natural; 

   function Processor_Utilization_Over_Deadline (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double; 


   function Processor_Utilization_Over_Period (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ) 
     return Double; 


   procedure Bound_On_Processor_Utilization (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String;        
         Result         : in out Double;                  
         Msg            : in out Unbounded_String         ); 


   procedure Dispatched_Bound_On_Processor_Utilization (
         My_Scheduler   : in     Generic_Scheduler; 
         My_Tasks       : in     Tasks_Set;         
         Processor_Name : in     Unbounded_String;  
         Result         : in out Double;            
         Msg            : in out Unbounded_String   ) is 
   abstract;


   procedure Dispatched_Response_Time (
         My_Scheduler   : in     Generic_Scheduler; 
         My_Tasks       : in out Tasks_Set;         
         Processor_Name : in     Unbounded_String;  
         Msg            : in out Unbounded_String;  
         Response_Time  :    out Response_Table     ) is 
   abstract;



   procedure Dispatched_Reset (
         A_Scheduler : in out Generic_Scheduler ) is 
   abstract;


   procedure Dump (
         My_Scheduler : in     Generic_Scheduler'Class ); 

   --
   --  Procedure and functions for scheduling
   --

   function Build_Tcb (
         My_Scheduler : in     Generic_Scheduler'Class; 
         A_Task       :        Task_Ptr                 ) 
     return Tcb_Ptr; 

   procedure Check_Before_Scheduling (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ); 

   procedure Do_Election (
         My_Scheduler : in     Generic_Scheduler'Class; 
         A_Tcb        : in     Tcb_Table;               
         At_Time      :        Natural;                 
         Elected      : in out Tasks_Range;             
         No_Task      : in out Boolean                  ); 

   function Dispatched_Build_Tcb (
         My_Scheduler : in     Generic_Scheduler; 
         A_Task       :        Task_Ptr           ) 
     return Tcb_Ptr is 
   abstract;

   procedure Dispatched_Check_Before_Scheduling (
         My_Scheduler   : in     Generic_Scheduler; 
         My_Tasks       : in     Tasks_Set;         
         Processor_Name : in     Unbounded_String   ) is 
   abstract;

   procedure Dispatched_Do_Election (
         My_Scheduler : in     Generic_Scheduler; 
         A_Tcb        : in     Tcb_Table;         
         At_Time      :        Natural;           
         Elected      : in out Tasks_Range;       
         No_Task      : in out Boolean            ) is 
   abstract;


   function Check_Offset (
         A_Tcb   : Tcb_Ptr; 
         At_Time : Natural  ) 
     return Boolean; 


   procedure Periodic_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ); 

   procedure Priority_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ); 

   procedure Start_Time_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ); 

   procedure Offset_Control (
         My_Scheduler   : in     Generic_Scheduler'Class; 
         My_Tasks       : in     Tasks_Set;               
         Processor_Name : in     Unbounded_String         ); 


private

   type Generic_Scheduler is abstract new Ada.Finalization.Controlled with 
      record 
         Is_Preemptive : Preemptive_Type;  
         Name          : Scheduler_Type;  
      end record; 

end Scheduler ;


