------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Text_Io;
use Text_Io;

with Ada.Sequential_Io;

with Dependencies;
use Dependencies;
with Buffers;
use Buffers;
with Messages;
use Messages;



with Fixed_Strings;
use Fixed_Strings;


package body Task_Dependencies is


   procedure Reset (
         My_Dependencies : in out Tasks_Dependencies ) is 

      Root_Task : Task_Ptr;  
      Empty_Set : Depends_Set;  

   begin

      My_Dependencies.Number_Of_Tasks:=0;

      Root_Task := new Periodic_Task;
      Root_Task.Name:=To_Unbounded_String("Root");

      My_Dependencies.Task_Position(Root_Index):=Root_Task;

      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            My_Dependencies.Dependencies(I,J):=Empty_Set;
         end loop;
      end loop;

   end Reset;



   function Search_Position (
         My_Dependencies : in     Tasks_Dependencies; 
         Name            : in     Unbounded_String    ) 
     return Dependencies_Range is 


   begin
      for I in  0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (Name = My_Dependencies.Task_Position(I).Name)
               then
            return I;
         end if;
      end loop;
      raise Task_Not_Found;
   end Search_Position;


   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean is 

      A_Task : Dependencies_Range;  
      Result : Boolean            := False;  
   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(I, A_Task)))
               then
            Result:=True;
         end if;
      end loop;
      return Result;
   end Has_Predecessor;


   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean is 

      A_Task : Dependencies_Range;  
      Result : Boolean            := False;  
   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(A_Task, I)))
               then
            Result:=True;
         end if;
      end loop;
      return Result;

   end Has_Successor;


   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set is 

      Result : Tasks_Set;  
      A_Task : Dependencies_Range;  

   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(A_Task, I)))
               then
            Add(Result, My_Dependencies.Task_Position(I));
         end if;
      end loop;
      return Result;

   end Get_Successors_List;


   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set is 

      Result : Tasks_Set;  
      A_Task : Dependencies_Range;  

   begin

      A_Task:=Search_Position(My_Dependencies, My_Task);

      for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if (not Is_Empty(My_Dependencies.Dependencies(I, A_Task)))
               then
            Add(Result, My_Dependencies.Task_Position(I));
         end if;
      end loop;
      return Result;

   end Get_Predecessors_List;


   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set is 

   begin
      return Get_Successors_List(My_Dependencies, My_Task.Name);
   end Get_Successors_List;


   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set is 

   begin
      return Get_Predecessors_List(My_Dependencies, My_Task.Name);
   end Get_Predecessors_List;


   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean is 

   begin
      return Has_Predecessor(My_Dependencies, My_Task.Name);
   end Has_Predecessor;


   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean is 

   begin
      return Has_Successor(My_Dependencies, My_Task.Name);
   end Has_Successor;


   procedure Dump (
         My_Dependencies : in     Tasks_Dependencies ) is 
   begin

      if My_Dependencies.Number_Of_Tasks /= 0
            then
         Put("Task list/index : ");
         New_Line;
         for I in 0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
            Put(" Task " & To_String(My_Dependencies.Task_Position(I).
                  Name) &
               " index is " & I'Img);
            New_Line;
         end loop;

         Put(" Task " & To_String(My_Dependencies.Task_Position(
                  Root_Index).
               Name) &
            " index is " & Root_Index'Img);
         New_Line;


         Put("Dependencies : ");
         New_Line;
         for I in  Dependencies_Range loop
            for J in  Dependencies_Range loop

               if (not Is_Empty(My_Dependencies.Dependencies(I, J)))
                     then
                  Put(To_String(My_Dependencies.Task_Position(I).Name) &
                     " => "
                     & To_String(My_Dependencies.Task_Position(J).Name));
                  New_Line;
               end if;

            end loop;
         end loop;

      end if;
   end Dump;



   function No_Dependency (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean is 

      Has_Dependency : Boolean := False;  

   begin
      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            if Is_Empty(My_Dependencies.Dependencies(I, J))
                  then
               Has_Dependency:=True;
            end if;

         end loop;
      end loop;

      return Has_Dependency;

   end No_Dependency;


   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr;           
         My_Deps         : in     Dependency_Ptr      ) is 
   begin
      Add_Tasks_Dependencies(My_Dependencies, My_Tasks,
         From_Task.Name, To_Task.Name, My_Deps);
   end Add_Tasks_Dependencies;


   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ) is 
   begin
      Delete_Tasks_Dependencies(My_Dependencies,
         From_Task.Name, To_Task.Name);
   end Delete_Tasks_Dependencies;


   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ) is 

      I,  
      J         : Dependencies_Range;  
      Empty_Set : Depends_Set;  
   begin

      I:=Search_Position(My_Dependencies,From_Task);
      J:=Search_Position(My_Dependencies,To_Task);

      My_Dependencies.Dependencies(I,J):=Empty_Set;
   end Delete_Tasks_Dependencies;


   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String;   
         My_Deps         : in     Dependency_Ptr      ) is 

      Found_To    : Boolean            := False;  
      Found_From  : Boolean            := False;  
      I,  
      J           : Dependencies_Range;  
      My_Iterator : Tasks_Iterator;  
      A_Task      : Task_Ptr;  
   begin

      -- Check/insert task name
      --
      for I in  0 .. (My_Dependencies.Number_Of_Tasks - 1) loop
         if(My_Dependencies.Task_Position(I).Name = From_Task)
               then
            Found_From:=True;
         end if;
         if(My_Dependencies.Task_Position(I).Name = To_Task)
               then
            Found_To:=True;
         end if;
      end loop;


      if (not Found_To)
            then
         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Tasks, A_Task, My_Iterator);

            exit when To_Task = A_Task.Name;

            if (Is_Last_Element(My_Tasks, My_Iterator) = True)
                  then
               raise Task_Not_Found;
            end if;

            Next_Element(My_Tasks, My_Iterator);


         end loop;

         My_Dependencies.Task_Position(My_Dependencies.Number_Of_Tasks):=
            A_Task;
         My_Dependencies.Number_Of_Tasks:=My_Dependencies.Number_Of_Tasks +
            1;
      end if;

      if (not Found_From)
            then
         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Tasks,A_Task,My_Iterator);

            exit when From_Task = A_Task.Name;

            if (Is_Last_Element(My_Tasks, My_Iterator) = True)
                  then
               raise Task_Not_Found;
            end if;

            Next_Element(My_Tasks, My_Iterator);

         end loop;

         My_Dependencies.Task_Position(My_Dependencies.Number_Of_Tasks) :=
            A_Task;
         My_Dependencies.Number_Of_Tasks := My_Dependencies.
            Number_Of_Tasks +
            1;
      end if;

      -- Add dependencies

      I:=Search_Position(My_Dependencies,From_Task);
      J:=Search_Position(My_Dependencies,To_Task);



      if (not Is_Empty(My_Dependencies.Dependencies(J,I)))
            then
         raise Cyclic_Dependencies_Error;
      end if;

      Add(My_Dependencies.Dependencies(I,J),My_Deps);
   end Add_Tasks_Dependencies;



   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ) 
     return Depends_Set is 

      A_Task1,  
      A_Task2 : Dependencies_Range;  

   begin

      A_Task1:=Search_Position(My_Dependencies, From_Task);
      A_Task2:=Search_Position(My_Dependencies, To_Task);

      if (not Is_Empty(My_Dependencies.Dependencies(A_Task1, A_Task2)))
            then
         return My_Dependencies.Dependencies(A_Task1, A_Task2);
      else
         raise No_Dependencies;
      end if;
   end Get_Dependencies;


   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ) 
     return Depends_Set is 
   begin
      return Get_Dependencies(My_Dependencies, From_Task.Name, To_Task.
         Name);
   end Get_Dependencies;




   procedure Fill_Descriptor (
         My_Deps : in     Depends_Set;      
         Desc    :    out Dummy_Descriptor; 
         A_Dep   : in     Dependency_Ptr    ) is 
   begin
      Desc:=0;
   end Fill_Descriptor;


   procedure Extract_Descriptor (
         My_Deps : in     Depends_Set;      
         Desc    : in     Dummy_Descriptor; 
         A_Dep   :    out Dependency_Ptr    ) is 
   begin
      A_Dep:=null;
   end Extract_Descriptor;




   type Tasks_Table_Name is array (Dependencies_Range) of Unbounded_String; 


   type File_Descriptor is 
      record 
         From_Task : Fixed_String;  
         To_Task   : Fixed_String;  
         Dep_Type  : Dependency_Type;  
         Dep_Name  : Fixed_String;  
      end record; 


   package Dep_File is new Ada.Sequential_Io( File_Descriptor );
   use Dep_File;


   procedure Write_To_File (
         My_Dependencies : in     Tasks_Dependencies;      
         File_Name       : in     Unbounded_String;        
         Write_Mode      : in     Io_Definitions.File_Mode ) is 

      Desc        : File_Descriptor;  
      A_Dep       : Dependency_Ptr;  
      My_Iterator : Depends_Iterator;  
      A_Dep_File  : Dep_File.File_Type;  
   begin

      if(Write_Mode = Save_In_New_File)
            then
         Create(A_Dep_File,
            Mode => Out_File,
            Name => To_String(File_Name));
      else
         Open(A_Dep_File,
            Mode => Append_File,
            Name => To_String(File_Name));
      end if;

      if not Is_Open(A_Dep_File)
            then
         raise Can_Not_Open_File_For_Writing;
      end if;


      for I in  Dependencies_Range loop
         for J in  Dependencies_Range loop

            if (not Is_Empty(My_Dependencies.Dependencies(I, J)))
                  then

               Reset_Iterator(My_Iterator);

               loop
                  Current_Element(My_Dependencies.Dependencies(I, J)
                     ,A_Dep,My_Iterator);

                  Desc.From_Task:=To_Bounded_String(To_String(
                        My_Dependencies.Task_Position(I).Name));
                  Desc.To_Task:=To_Bounded_String(To_String(
                        My_Dependencies.Task_Position(J).Name));

                  Desc.Dep_Type:=A_Dep.Dependence_Type;

                  if A_Dep.Dependence_Type=Precedence_Dependency
                        then
                     Desc.Dep_Name:=To_Bounded_String("");
                  else
                     if A_Dep.Dependence_Type=Buffer_Dependency
                           then
                        Desc.Dep_Name:=To_Bounded_String(To_String(
                              Buffer_Ptr(A_Dep).Name));
                     else
                        Desc.Dep_Name:=To_Bounded_String(To_String(
                              Message_Ptr(A_Dep).Name));
                     end if;
                  end if;

                  Write(A_Dep_File,Desc);

                  exit when Is_Last_Element(My_Dependencies.Dependencies(
                        I, J),
                     My_Iterator) = True;

                  Next_Element(My_Dependencies.Dependencies(I, J),
                     My_Iterator);

               end loop;

            end if;

         end loop;
      end loop;


      Close(A_Dep_File);
   end Write_To_File;



   procedure Read_From_File (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         My_Messages     : in     Messages_Set;       
         My_Buffers      : in     Buffers_Set;        
         File_Name       : in     Unbounded_String    ) is 

      Desc       : File_Descriptor;  
      A_Dep_File : Dep_File.File_Type;  

      A_Pred        : Precedence_Ptr;  
      A_Buff        : Buffer_Ptr;  
      A_Mess        : Message_Ptr;  
      Buff_Iterator : Buffers_Iterator;  
      Mess_Iterator : Messages_Iterator;  

   begin

      Reset(My_Dependencies);

      Open(A_Dep_File,
         Mode => In_File,
         Name => To_String(File_Name));


      if not Is_Open(A_Dep_File)
            then
         raise Can_Not_Open_File_For_Reading;
      end if;


      while not End_Of_File(A_Dep_File)
            loop
         Read(A_Dep_File,Desc);

         if Desc.Dep_Type=Precedence_Dependency
               then
            A_Pred := new Precedence;
            Add_Tasks_Dependencies(My_Dependencies, My_Tasks,
               To_Unbounded_String(To_String(Desc.From_Task)),
               To_Unbounded_String(To_String(Desc.To_Task)),
               Dependency_Ptr(A_Pred) );
         else
            if Desc.Dep_Type=Buffer_Dependency
                  then

               Reset_Iterator(Buff_Iterator);

               loop
                  Current_Element(My_Buffers, A_Buff ,Buff_Iterator);

                  if A_Buff.Name = To_Unbounded_String(To_String(Desc.
                           Dep_Name))
                        then
                     Add_Tasks_Dependencies(My_Dependencies, My_Tasks,
                        To_Unbounded_String(To_String(Desc.From_Task)),
                        To_Unbounded_String(To_String(Desc.To_Task)),
                        Dependency_Ptr(A_Buff) );
                     exit;
                  end if;

                  if Is_Last_Element(My_Buffers, Buff_Iterator) = True
                        then
                     raise Buffer_Not_Found;
                  end if;

                  Next_Element(My_Buffers, Buff_Iterator);

               end loop;


            else

               Reset_Iterator(Mess_Iterator);

               loop
                  Current_Element(My_Messages, A_Mess, Mess_Iterator);

                  if A_Mess.Name = To_Unbounded_String(To_String(Desc.
                           Dep_Name))
                        then
                     Add_Tasks_Dependencies(My_Dependencies, My_Tasks,
                        To_Unbounded_String(To_String(Desc.From_Task)),
                        To_Unbounded_String(To_String(Desc.To_Task)),
                        Dependency_Ptr(A_Mess) );
                     exit;
                  end if;

                  if Is_Last_Element(My_Messages, Mess_Iterator) = True
                        then
                     raise Message_Not_Found;
                  end if;

                  Next_Element(My_Messages, Mess_Iterator);

               end loop;

            end if;
         end if;

      end loop;

      Close(A_Dep_File);
   end Read_From_File;



   procedure Dump (
         My_Deps : in     Depends_Set ) is 

      A_Dep       : Dependency_Ptr;  
      My_Iterator : Depends_Iterator;  

   begin
      Put_Line ("Number of dependencies : " & Get_Number_Of_Elements(
            My_Deps)'
         Img);

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Deps,A_Dep,My_Iterator);
         Dump(A_Dep.all);
         New_Line;

         exit when Is_Last_Element(My_Deps, My_Iterator) = True;

         Next_Element(My_Deps, My_Iterator);

      end loop;
   end Dump;




end Task_Dependencies;


