------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Config;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;

with Tasks;
use Tasks;

with Task_Set;
use Task_Set;
use Task_Set.Generic_Task_Set;
with Buffer_Set;
use Buffer_Set;
use Buffer_Set.Generic_Buffer_Set;
with Message_Set;
use Message_Set;
use Message_Set.Generic_Message_Set;

with Dependencies;
use Dependencies;

with Io_Definitions;
use Io_Definitions;

with Sets;


package Task_Dependencies is





   type Tasks_Dependencies is private; 



   type Dummy_Descriptor is new Integer; 

   package Dep_Set is new Sets (
      Max_Element     => Config.Max_Tasks, 
      File_Descriptor => Dummy_Descriptor, 
      Element         => Dependency_Ptr);
   use Dep_Set;

   type Depends_Set is new Dep_Set.Set with private; 

   subtype Depends_Range is Dep_Set.Element_Range;
   subtype Depends_Iterator is Dep_Set.Iterator;

   procedure Dump (
         My_Deps : in     Depends_Set ); 


   procedure Fill_Descriptor (
         My_Deps : in     Depends_Set;      
         Desc    :    out Dummy_Descriptor; 
         A_Dep   : in     Dependency_Ptr    ); 


   procedure Extract_Descriptor (
         My_Deps : in     Depends_Set;      
         Desc    : in     Dummy_Descriptor; 
         A_Dep   :    out Dependency_Ptr    ); 




   Cyclic_Dependencies_Error : exception;  
   Task_Not_Found            : exception;  
   Message_Not_Found         : exception;  
   Buffer_Not_Found          : exception;  
   No_Dependencies           : exception;  


   procedure Reset (
         My_Dependencies : in out Tasks_Dependencies ); 



   procedure Read_From_File (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         My_Messages     : in     Messages_Set;       
         My_Buffers      : in     Buffers_Set;        
         File_Name       : in     Unbounded_String    ); 

   procedure Write_To_File (
         My_Dependencies : in     Tasks_Dependencies; 
         File_Name       : in     Unbounded_String;   
         Write_Mode      : in     File_Mode           ); 


   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set; 


   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Tasks_Set; 


   function Get_Successors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set; 


   function Get_Predecessors_List (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Tasks_Set; 


   function No_Dependency (
         My_Dependencies : in     Tasks_Dependencies ) 
     return Boolean; 


   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean; 

   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Task_Ptr            ) 
     return Boolean; 

   function Has_Predecessor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean; 

   function Has_Successor (
         My_Dependencies : in     Tasks_Dependencies; 
         My_Task         : in     Unbounded_String    ) 
     return Boolean; 


   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ) 
     return Depends_Set; 

   function Get_Dependencies (
         My_Dependencies : in     Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ) 
     return Depends_Set; 


   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String    ); 

   procedure Delete_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr            ); 



   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task       : in     Task_Ptr;           
         To_Task         : in     Task_Ptr;           
         My_Deps         : in     Dependency_Ptr      ); 

   procedure Add_Tasks_Dependencies (
         My_Dependencies : in out Tasks_Dependencies; 
         My_Tasks        : in     Tasks_Set;          
         From_Task       : in     Unbounded_String;   
         To_Task         : in     Unbounded_String;   
         My_Deps         : in     Dependency_Ptr      ); 



   procedure Dump (
         My_Dependencies : in     Tasks_Dependencies ); 



private


   type Depends_Set is new Dep_Set.Set with null record; 

   Root_Index : constant := Config.Max_Tasks + 1;  
   type Dependencies_Range is new Natural range 0 .. Root_Index; 

   type Tasks_Table is array (Dependencies_Range) of Task_Ptr; 
   type Dependencies_Table is array (Dependencies_Range, Dependencies_Range) of Depends_Set; 


   type Tasks_Dependencies is 
      record 
         Dependencies    : Dependencies_Table;  
         Task_Position   : Tasks_Table;  
         Number_Of_Tasks : Dependencies_Range := 0;  
      end record; 


end Task_Dependencies;


