------------------------------------------------------------------------------
------------------------------------------------------------------------------
-- Cheddar is a real time simulator mainly designed for educational purpose. 
-- This program provides services to automaticly check temporal constraints 
-- of real time tasks.
--
-- Copyright (C) 2002   Frank Singhoff, Jerome Legrand
-- Cheddar is developped by the LIMI Team/EA 2215
--
-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
--
--
-- Contact : F. Singhoff (singhoff@univ-brest.fr)
--           J. Legrand (jerome.legrand@univ-brest.fr)
------------------------------------------------------------------------------
------------------------------------------------------------------------------





with Text_Io;
use Text_Io;

with Ada.Strings.Unbounded;
use Ada.Strings.Unbounded;
with Unbounded_Strings;
use Unbounded_Strings;

with Translate;
use Translate;


with Time_Units;
use  Time_Units;
use  Time_Units.A_Time_Unit;


use Task_Set.Generic_Task_Set;


package body Task_Set is



   procedure Fill_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     :    out Task_Descriptor; 
         A_Task   : in     Task_Ptr         ) is 

   begin
      Desc.Offset:=A_Task.Offset;
      Desc.Name := To_Bounded_String(To_String(A_Task.Name));
      Desc.Location:=To_Bounded_String(To_String(A_Task.Location));
      Desc.Capacity := A_Task.Capacity;
      Desc.Start_Time := A_Task.Start_Time;
      Desc.Priority:= A_Task.Priority;
      Desc.Deadline := A_Task.Deadline;
      Desc.Blocking_Time:= A_Task.Blocking_Time;

      Desc.Period:=0;
      Desc.Jitter:=0;

      if (A_Task.Task_Type = Periodic_Type)
            then
         Desc.Period := Periodic_Task_Ptr(A_Task).Period;
         Desc.Jitter := Periodic_Task_Ptr(A_Task).Jitter;
      end if;

   end Fill_Descriptor;


   procedure Extract_Descriptor (
         My_Tasks : in     Tasks_Set;       
         Desc     : in     Task_Descriptor; 
         A_Task   :    out Task_Ptr         ) is 

      New_Periodic_Task  : Periodic_Task_Ptr;  
      New_Aperiodic_Task : Aperiodic_Task_Ptr;  

   begin
      if (Desc.Period /= 0)
            then
         New_Periodic_Task:=new Periodic_Task;
         New_Periodic_Task.Jitter:=Desc.Jitter;
         New_Periodic_Task.Period:=Desc.Period;
         A_Task:=Task_Ptr(New_Periodic_Task);

      else
         New_Aperiodic_Task:=new Aperiodic_Task;
         A_Task:=Task_Ptr(New_Aperiodic_Task);
      end if;

      A_Task.Offset:=Desc.Offset;
      A_Task.Deadline:=Desc.Deadline;
      A_Task.Name:=To_Unbounded_String(To_String(Desc.Name));
      A_Task.Capacity:=Desc.Capacity;
      A_Task.Start_Time:=Desc.Start_Time;
      A_Task.Location:=To_Unbounded_String(To_String(Desc.Location));
      A_Task.Priority:=Desc.Priority;
      A_Task.Blocking_Time:=Desc.Blocking_Time;

   end Extract_Descriptor;

   procedure Add_Task (
         My_Tasks      : in out Tasks_Set;        
         Name          : in     Unbounded_String; 
         Location      : in     Unbounded_String; 
         Start_Time    : in     Natural;          
         Capacity      : in     Natural;          
         Period        : in     Natural;          
         Deadline      : in     Natural;          
         Jitter        : in     Natural;          
         Blocking_Time : in     Natural;          
         Priority      : in     Priority_Range;   
         Offset        :        Offset_Table      ) is 
      Dummy : Task_Ptr;  
   begin
      Add_Task(My_Tasks, Dummy, Name, Location, Start_Time, Capacity,
         Period,
         Deadline, Jitter, Blocking_Time, Priority, Offset);
   end Add_Task;

   procedure Add_Task (
         My_Tasks      : in out Tasks_Set;        
         A_Task        : in out Task_Ptr;         
         Name          : in     Unbounded_String; 
         Location      : in     Unbounded_String; 
         Start_Time    : in     Natural;          
         Capacity      : in     Natural;          
         Period        : in     Natural;          
         Deadline      : in     Natural;          
         Jitter        : in     Natural;          
         Blocking_Time : in     Natural;          
         Priority      : in     Priority_Range;   
         Offset        : in     Offset_Table      ) is 

      My_Iterator : Iterator;  

      Desc : Task_Descriptor;  

   begin

      if (Name = "")
            then
         raise Invalid_Parameter;
      end if;

      if (Location = "")
            then
         raise Invalid_Parameter;
      end if;

      if (Capacity = 0)
            then
         raise Invalid_Parameter;
      end if;

      if (Jitter > Deadline)
            then
         raise Invalid_Parameter;
      end if;

      if(Get_Number_Of_Elements(My_Tasks) > 0)
            then

         Reset_Iterator(My_Iterator);

         loop
            Current_Element(My_Tasks, A_Task, My_Iterator);
            if (Name = A_Task.Name)
                  then
               raise Task_Already_Defined;
            end if;

            exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

            Next_Element(My_Tasks, My_Iterator);
         end loop;
      end if;


      Desc.Offset:=Offset;
      Desc.Name:=To_Bounded_String(To_String(Name));
      Desc.Location:=To_Bounded_String(To_String(Location));
      Desc.Start_Time:=Start_Time;
      Desc.Capacity:=Capacity;
      Desc.Period:=Period;
      Desc.Deadline:=Deadline;
      Desc.Jitter:=Jitter;
      Desc.Priority:=Priority;
      Desc.Blocking_Time:=Blocking_Time;

      Extract_Descriptor(My_Tasks, Desc, A_Task);


      Add(My_Tasks,A_Task);

   end Add_Task;




   function Have_Deadlines_Equal_Than_Periods (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Boolean is 

      Equal_Than  : Boolean  := True;  
      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if ( (A_Task.Task_Type /= Periodic_Type) and
               (A_Task.Location = Processor_Name) )
               then
            raise Task_Must_Be_Periodic;
         end if;

         if   (A_Task.Location = Processor_Name)
               then
            if (A_Task.Deadline /=
                  Periodic_Task_Ptr(A_Task).Period)
                  then
               Equal_Than := False;
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return Equal_Than;

   end Have_Deadlines_Equal_Than_Periods ;




   procedure Dump (
         My_Tasks : in     Tasks_Set ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin
      Put_Line ("Number of tasks : " & Get_Number_Of_Elements(My_Tasks)'
         Img);

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);
         Dump(A_Task.all);
         New_Line;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;
   end Dump;


   function Search_Task (
         My_Tasks : in     Tasks_Set;       
         Name     : in     Unbounded_String ) 
     return Task_Ptr is 

      My_Iterator : Iterator;  
      A_Task      : Task_Ptr;  
      Result      : Task_Ptr;  

      Found : Boolean := False;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if(A_Task.Name = Name)
               then
            Found:=True;
            Result:=A_Task;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      if not Found
            then
         raise Task_Not_Found;
      end if;

      return Result;
   end Search_Task;


   procedure Check_Temporal_Constraints (
         My_Task : in     Periodic_Task_Ptr;   
         Sched   : in     Scheduling_Sequence; 
         Result  : in out Unbounded_String     ) is 

      Capacity         : Natural := My_Task.Capacity;  
      Dynamic_Deadline : Natural := 0;  
      Static_Deadline  : Natural := 0;  
      Temp             : Natural := 0;  
   begin
      if My_Task.Deadline=0
            then
         Static_Deadline:=My_Task.Period;
      else
         Static_Deadline:=My_Task.Deadline;
      end if;

      Dynamic_Deadline:=Static_Deadline+My_Task.Start_Time;

      for I in 0 .. Sched.Number_Of_Time_Unit-1 loop

         if (Sched.Scheduled(I).Task_Item.Name = My_Task.Name) then
            Capacity := Capacity - 1;
            if (Capacity = 0)  then
               if (Sched.Scheduled(I).Data+1 > Dynamic_Deadline)
                     then
                  Temp:=Sched.Scheduled(I).Data+1;
                  Result := Result & Lb_Periodic(Current_Language)
                     & To_String(My_Task.Name)
                     & Lb_Check_Deadline(Current_Language)
                     & My_Task.Deadline'Img
                     & Lb_Completion_Time(Current_Language)
                     & Temp'Img & To_Unbounded_String(")") & Unbounded_Lf;
               end if;
               Capacity:=My_Task.Capacity;
               Dynamic_Deadline:=Dynamic_Deadline+My_Task.Period;
            end if;
         end if;
      end loop;

   end Check_Temporal_Constraints;


   procedure Check_Temporal_Constraints (
         My_Task : in     Aperiodic_Task_Ptr;  
         Sched   : in     Scheduling_Sequence; 
         Result  : in out Unbounded_String     ) is 

      Capacity : Natural := My_Task.Capacity;  
      Temp     : Natural := 0;  
   begin

      for I in 0 .. Sched.Number_Of_Time_Unit-1 loop

         if (Sched.Scheduled(I).Task_Item.Name = My_Task.Name) then
            Capacity := Capacity - 1;
            if (Capacity = 0)  then
               if (Sched.Scheduled(I).Data+1 > My_Task.Deadline)
                     then
                  Temp:=Sched.Scheduled(I).Data+1;
                  Result := Result & Lb_Aperiodic(Current_Language)
                     & To_String(My_Task.Name)
                     & Lb_Check_Deadline(Current_Language)
                     & My_Task.Deadline'Img
                     & Lb_Completion_Time(Current_Language)
                     & Temp'Img & To_Unbounded_String(")") & Unbounded_Lf;
               end if;
            end if;
         end if;
      end loop;

   end Check_Temporal_Constraints;



   procedure Check_Temporal_Constraints (
         My_Tasks  : in     Tasks_Set;           
         Sched     : in     Scheduling_Sequence; 
         Processor :        Unbounded_String;    
         Result    : in out Unbounded_String     ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      Result:=To_Unbounded_String("");
      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (Processor=A_Task.Location)
               then
            if A_Task.Task_Type = Periodic_Type
                  then
               Check_Temporal_Constraints(Periodic_Task_Ptr(A_Task),Sched,
                  Result);
            else
               Check_Temporal_Constraints(Aperiodic_Task_Ptr(A_Task),
                  Sched,Result);
            end if;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

   end Check_Temporal_Constraints;

   function Increasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Priority<= Op2.Priority);
   end Increasing_Priority;

   function Decreasing_Priority (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Priority>= Op2.Priority);
   end Decreasing_Priority;


   function Increasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Periodic_Task_Ptr(Op1).Period <= Periodic_Task_Ptr(Op2).
         Period);
   end Increasing_Period;


   function Decreasing_Period (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Periodic_Task_Ptr(Op1).Period >= Periodic_Task_Ptr(Op2).
         Period);
   end Decreasing_Period;


   function Increasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Deadline <= Op2.Deadline);
   end Increasing_Deadline;


   function Decreasing_Deadline (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Deadline >= Op2.Deadline);
   end Decreasing_Deadline;

   function Increasing_Name (
         Op1 : in     Task_Ptr; 
         Op2 : in     Task_Ptr  ) 
     return Boolean is 
   begin
      return (Op1.Name<=Op2.Name);
   end Increasing_Name;



   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Unbounded_String is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Location)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return A_Task.Location;

   end Get;


   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Priority_Range is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Priority)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return A_Task.Priority;

   end Get;


   function Get (
         My_Tasks   : in     Tasks_Set;        
         Task_Name  : in     Unbounded_String; 
         Param_Name : in     Task_Parameters   ) 
     return Natural is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( (Param_Name /= Start_Time) and
            (Param_Name /= Blocking_Time) and
            (Param_Name /= Capacity) and
            (Param_Name /= Period) and
            (Param_Name /= Deadline) and
            (Param_Name /= Jitter) )
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      if (Param_Name = Start_Time)
            then
         return A_Task.Start_Time;
      else
         if (Param_Name = Deadline)
               then
            return A_Task.Deadline;
         else
            if (Param_Name = Period)
                  then
               return Periodic_Task_Ptr(A_Task).Period;
            else
               if (Param_Name = Jitter)
                     then
                  return Periodic_Task_Ptr(A_Task).Jitter;
               else
                  if (Param_Name = Capacity)
                        then
                     return A_Task.Capacity;
                  else
                     return A_Task.Blocking_Time;
                  end if;
               end if;
            end if;
         end if;
      end if;

   end Get;


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Unbounded_String  ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Location)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            A_Task.Location:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

   end Set;


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Priority_Range    ) is 

      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if (Param_Name /= Priority)
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            A_Task.Priority:=Param_Value;
            exit;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;
   end Set;


   procedure Set (
         My_Tasks    : in out Tasks_Set;        
         Task_Name   : in     Unbounded_String; 
         Param_Name  : in     Task_Parameters;  
         Param_Value : in     Natural           ) is 
      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      if ( (Param_Name /= Start_Time) and
            (Param_Name /= Blocking_Time) and
            (Param_Name /= Capacity) and
            (Param_Name /= Period) and
            (Param_Name /= Deadline) and
            (Param_Name /= Jitter) )
            then
         raise Invalid_Parameter;
      end if;

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Name = Task_Name)
               then
            if (Param_Name = Start_Time)
                  then
               A_Task.Start_Time:=Param_Value;
               exit;
            end if;
            if (Param_Name = Deadline)
                  then
               A_Task.Deadline:=Param_Value;
            end if;
            if (Param_Name = Period)
                  then
               Periodic_Task_Ptr(A_Task).Period:=Param_Value;
               exit;
            end if;
            if (Param_Name = Jitter)
                  then
               Periodic_Task_Ptr(A_Task).Jitter:=Param_Value;
               exit;
            end if;
            if (Param_Name = Capacity)
                  then
               A_Task.Capacity:=Param_Value;
               exit;
            end if;
            if (Param_Name = Blocking_Time)
                  then
               A_Task.Blocking_Time:=Param_Value;
               exit;
            end if;
         end if;
         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;
   end Set;


   function Get_Number_Of_Task_From_Processor (
         My_Tasks       : in     Tasks_Set;       
         Processor_Name : in     Unbounded_String ) 
     return Tasks_Range is 

      Number      : Tasks_Range := 0;  
      A_Task      : Task_Ptr;  
      My_Iterator : Iterator;  

   begin

      Reset_Iterator(My_Iterator);

      loop
         Current_Element(My_Tasks,A_Task,My_Iterator);

         if (A_Task.Location  = Processor_Name)
               then
            Number:=Number+1;
         end if;

         exit when Is_Last_Element(My_Tasks, My_Iterator) = True;

         Next_Element(My_Tasks, My_Iterator);

      end loop;

      return Number;
   end Get_Number_Of_Task_From_Processor;



end Task_Set;
